/// ~english
/// The presence property class that contains presence properties, including the publisher's user ID and current presence state, and the platform used by the online device, as well as the presence's extension information, update time, and subscription expiration time.
/// ~end
///
/// ~chinese
/// 在线状态属性类，包含发布者的用户名、在线设备使用的平台、当前在线状态以及在线状态的扩展信息、更新时间和到期时间。
/// ~end
class EMPresence {
  /// ~english
  /// The user ID of the presence publisher.
  /// ~end
  ///
  /// ~chinese
  /// 在线状态发布者的用户 ID。
  /// ~end
  final String publisher;

  /// ~english
  /// The presence description information.
  /// ~end
  ///
  /// ~chinese
  /// 自定义在线状态，例如忙碌、离开和隐身等。
  /// ~end
  final String statusDescription;

  /// ~english
  /// The presence update time, which is generated by the server.
  /// ~end
  ///
  /// ~chinese
  /// 在线状态更新 Unix 时间戳，单位为秒。
  /// ~end
  final int lastTime;

  /// ~english
  /// The expiration time of the presence subscription.
  /// ~end
  ///
  /// ~chinese
  /// 在线状态订阅到期 Unix 时间戳，单位为秒。
  /// ~end
  final int expiryTime;

  /// ~english
  /// The details of the current presence state.
  /// ~end
  ///
  /// ~chinese
  /// 该用户的当前在线状态详情。
  /// ~end
  Map<String, int>? statusDetails;

  /// @nodoc
  EMPresence._private(
    this.publisher,
    this.statusDescription,
    this.statusDetails,
    this.lastTime,
    this.expiryTime,
  );

  /// @nodoc
  factory EMPresence.fromJson(Map map) {
    String publisher = map["publisher"] ?? "";
    String statusDescription = map["statusDescription"] ?? "";
    int latestTime = map["lastTime"] ?? 0;
    int expiryTime = map["expiryTime"] ?? 0;
    Map<String, int>? statusDetails = map["statusDetails"]?.cast<String, int>();
    return EMPresence._private(
        publisher, statusDescription, statusDetails, latestTime, expiryTime);
  }
}

/// ~english
/// The presence details, including the platform used by the publisher's current online device and the current presence state.
/// ~end
///
/// ~chinese
/// 用户在线状态详情。
/// ~end
class EMPresenceStatusDetail {
  /// ~english
  /// The platform used by the current online device of the publisher, which can be "ios", "android", "linux", "win", or "webim".
  /// ~end
  ///
  /// ~chinese
  /// 发布在线设备，可能是 "ios", "android", "linux", "win", "webim"。
  /// ~end
  final String device;

  /// ~english
  /// The current presence state of the publisher.
  /// ~end
  ///
  /// ~chinese
  /// 发布的在线状态。
  /// ~end
  final int status;

  /// @nodoc
  EMPresenceStatusDetail._private(
    this.device,
    this.status,
  );

  /// @nodoc
  factory EMPresenceStatusDetail.fromJson(Map map) {
    String device = map["device"];
    int status = map["status"] ?? 0;
    return EMPresenceStatusDetail._private(device, status);
  }
}
